/**CFile***********************************************************************

  FileName    [mcInvar.c]

  PackageName [mc]

  Synopsis    [Dedicated algorithms for the verification of
  invariants on-the-fly wrt reachability analysis.]

  Description [Dedicated algorithms for the verification of
  invariants on-the-fly wrt reachability analysis.]

  SeeAlso     [mcMc.c]

  Author      [Marco Roveri]

  Copyright   [ Copyright (c) 1998 by ITC-IRST and Carnegie Mellon University.
  All Rights Reserved. This software is for educational purposes only.
  Permission is given to academic institutions to use, copy, and modify
  this software and its documentation provided that this introductory
  message is not removed, that this software and its documentation is
  used for the institutions' internal research and educational purposes,
  and that no monies are exchanged. No guarantee is expressed or implied
  by the distribution of this code.
  Send bug-reports and/or questions to: nusmv@irst.itc.it. ]

******************************************************************************/

#include "mcInt.h" 

static char rcsid[] UTIL_UNUSED = "$Id: $";

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

extern bdd_ptr state_variables_bdd;
extern bdd_ptr actions_bdd;
extern node_ptr all_symbols;

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/* Enable/Disable the forward/backward heuristic */
#define FB_HEURISTIC

/* The heuristic to use in the forward/backward search */
#define FORWARD_BACKWARD_HEURISTICS 0

/* Define it to 1 if you don't want the optimized image computation */
#define FORWARD_BACKWARD_NO_OPT 1

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Verifies that M,s0 |= AG alpha]

  Description [Verifies that M,s0 |= AG alpha, with alpha propositional.]

  SideEffects []

  SeeAlso     [check_spec check_ltlspec]

******************************************************************************/
void check_invar(node_ptr invar_expr_lst)
{
  /* node_ptr invar_cxt = car(invar_expr_lst); */
  node_ptr invar_expr = cdr(invar_expr_lst);

  fprintf(nusmv_stdout, "Checking invariant: ");
  print_node(nusmv_stdout, invar_expr);
  fprintf(nusmv_stdout, "\n");
  check_invariant_forward(invar_expr);
}

/**Function********************************************************************

  Synopsis    [Performs invariant checking going backward and forward.]

  Description [Performs invariant checking by alternating a step
  backward and a step forward.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
void check_invar_fb(node_ptr invar_expr_lst)
{
  node_ptr l;
  /* node_ptr invar_cxt = car(invar_expr_lst); */
  node_ptr invar_expr = cdr(invar_expr_lst);

  l = invar_expr;
  while (l) {
    node_ptr cur_inv = car(l);

    l = cdr(l);
    fprintf(nusmv_stdout, "Checking invariant: ");
    print_node(stdout, cur_inv);
#if FORWARD_BACKWARD_NO_OPT
    fprintf(nusmv_stdout, ".\n Using Forward and Backward Search.\n");
    check_invariant_forward_backward(cur_inv);
#else
    fprintf(nusmv_stdout, ".\n Using Forward and Backward OPT Search.\n");
    check_invariant_forward_backward_opt(cur_inv);
#endif
  }
}


/**Function********************************************************************

  Synopsis           [Extracts and prints a counterexample for AG alpha.]

  Description        [Extracts a counterexample that leads to a state
  not satisfying the invariant AG alpha. The counterexample produced
  is the shortest execution trace that exploits the falsity of the
  invariant. The computed counterexample is printed out.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void compute_and_print_path(bdd_ptr reached_goal_bdd, node_ptr path_slices, int nstep, int flag)
{
  int i;
  node_ptr path;
  bdd_ptr state;

  nstep = nstep - 1;
  state = bdd_pick_one_state_rand(dd_manager, reached_goal_bdd);
  path = cons((node_ptr)state, Nil);
  /*
    We ignore the first element of path_slices, which is the current frontier
    of reachable states 
  */
  if (path_slices != Nil) path_slices = cdr(path_slices);
  for(i = nstep; (i >= 0) ; i--) {
    bdd_ptr intersection, image;
    
    if (car(path_slices) == Nil) internal_error("path_slices == Nil: %d", i);
    image = Img_ImageBwd(state);
    intersection = bdd_and(dd_manager, image, (bdd_ptr)car(path_slices));
    if (bdd_is_zero(dd_manager, intersection)) {
      bdd_free(dd_manager, image);
      bdd_free(dd_manager, state);
      bdd_free(dd_manager, intersection);
      internal_error("Intersection == emptyset: Step %d",i);
    }
    state = bdd_pick_one_state(dd_manager, intersection);
    path = cons((node_ptr)state, path);
    bdd_free(dd_manager, image);
    bdd_free(dd_manager, intersection);
    path_slices = cdr(path_slices);
  }
  {
    node_ptr l = (flag == 0) ? path : reverse(path);

    print_explanation(path);

    /* free the list of path */
    l = path;
    while( l ) {
      node_ptr m = l;

      bdd_free(dd_manager,(bdd_ptr)car(l));
      l = cdr(l);
      free_node(m);
    }
  }
}

/**Function********************************************************************

  Synopsis           [Performs on the fly verification of the
  invariant during reachability analysis.]

  Description        [During the computation of reachable states it
  checks invariants. If the invariant is not satisfied, then an
  execution trace leading to a state not satisfing the invariant is
  printed out.]

  SideEffects        []

  SeeAlso            [check_invariant_forward_opt]

******************************************************************************/
void check_invariant_forward(node_ptr inv_expr)
{ 

  bdd_ptr reachable_states;
  int nstep = 0;
  node_ptr plan_reach_list = Nil;
  bdd_ptr target_states = bdd_zero(dd_manager);
  bdd_ptr f_frontier = bdd_and(dd_manager, init_bdd, invar_bdd);
  bdd_ptr tmp = eval_spec(inv_expr,Nil);
  bdd_ptr goal_states = bdd_not(dd_manager, tmp);
  bdd_ptr zero = bdd_zero(dd_manager);

  bdd_free(dd_manager, tmp);

  /* The goal states have to be intersected with the invariants */
  bdd_and_accumulate(dd_manager, &goal_states, invar_bdd);

  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr,"Size of invariant states: %d bdd nodes, %g states\n",
            bdd_size(dd_manager, goal_states), bdd_count_states(dd_manager, goal_states));
    fprintf(nusmv_stderr,"Size of initial states: %d bdd nodes\n", bdd_size(dd_manager, init_bdd));
    fprintf(nusmv_stderr,"Starting to slice the space of reachable states ...\n");
  }

  plan_reach_list = cons((node_ptr)bdd_dup(f_frontier), plan_reach_list);
  reachable_states = bdd_dup(init_bdd);
  while ((f_frontier != zero) && (target_states == zero)) {
    bdd_ptr f_prev_reachable = bdd_dup(reachable_states);
    
    if (opt_verbose_level_gt(options, 0))
      fprintf(nusmv_stderr, "iteration %d: BDD size: %d bdd nodes, states = %g; frontier size = %d bdd nodes.\n",
              nstep, bdd_size(dd_manager, reachable_states),
              bdd_count_states(dd_manager, reachable_states),
              bdd_size(dd_manager, f_frontier));
    
    /* Intersect the frontier with normal assignments */
    bdd_and_accumulate(dd_manager, &f_frontier, invar_bdd);
    if (opt_verbose_level_gt(options, 0))
      fprintf(nusmv_stderr,"invariant combined: %d bdd nodes\n",
              bdd_size(dd_manager, f_frontier));
    {
      bdd_ptr image = Img_ImageFwd(f_frontier);

      bdd_or_accumulate(dd_manager, &reachable_states, image);
      bdd_free(dd_manager, image);
    }
    if (opt_verbose_level_gt(options, 0))
      fprintf(nusmv_stderr, "forward step done: %d bdd nodes\n",
              bdd_size(dd_manager, reachable_states));

    {
      bdd_ptr not_f_prev_reachable = bdd_not(dd_manager, f_prev_reachable);
      
      bdd_free(dd_manager, f_frontier);
      f_frontier = bdd_and(dd_manager, reachable_states, not_f_prev_reachable);
      bdd_free(dd_manager, not_f_prev_reachable);
    }
    plan_reach_list = cons((node_ptr)bdd_dup(f_frontier), plan_reach_list);
    if (opt_verbose_level_gt(options, 0))
      fprintf(nusmv_stderr,"new frontier computed: %d bdd nodes\n",
              bdd_size(dd_manager, f_frontier));
    bdd_free(dd_manager, target_states);
    target_states = bdd_and(dd_manager, goal_states, f_frontier);
    if (opt_verbose_level_gt(options, 0)) {
      if (target_states != zero) {
        fprintf(nusmv_stderr,
                "States not satisfying the invariant reached at step %d: %d bdd nodes, %g states\n",
                nstep, bdd_size(dd_manager, target_states), bdd_count_states(dd_manager, target_states));
      }
    }
    bdd_free(dd_manager, f_prev_reachable);
    nstep++;
  }
  fprintf(nusmv_stdout, "-- invariant ");
  print_node(nusmv_stdout, inv_expr);
  if (target_states != zero) {
    fprintf(nusmv_stdout, " is false\n");
    compute_and_print_path(target_states, plan_reach_list, nstep, 0);
  }
  else {
    fprintf(nusmv_stdout, " is true\n");
  }
  bdd_free(dd_manager, target_states);
  bdd_free(dd_manager, f_frontier);
  bdd_free(dd_manager, zero);
  bdd_free(dd_manager, reachable_states);
  bdd_free(dd_manager, goal_states);
  walk_dd(dd_manager, bdd_free, plan_reach_list);
  {
    node_ptr l = plan_reach_list;

    while (l) {
      node_ptr m = l;

      l = cdr(l);
      free_node(m);
    }
  }
}

/**Function********************************************************************

  Synopsis           [Performs on the fly verification of the
  invariant during reachability analysis.]

  Description        [During the computation of reachable states it
  checks invariants. If the invariant is not satisfied, then an
  execution trace leading to a state not satisfing the invariant is
  printed out. It uses an optimized algorithm to perform image
  computation based on cofactoring. It works only with the monolithic
  transition relation.]

  SideEffects        []

  SeeAlso            [check_invariant_forward]

******************************************************************************/
void check_invariant_forward_opt(node_ptr inv_expr)
{ 
  int nstep = 0;
  node_ptr plan_reach_list = Nil;
  bdd_ptr reachable_states;
  bdd_ptr target_states = bdd_zero(dd_manager);
  bdd_ptr zero = bdd_zero(dd_manager);
  /* f_frontier is the frontier */
  bdd_ptr f_frontier = bdd_and(dd_manager, init_bdd, invar_bdd);
  bdd_ptr tmp = eval_spec(inv_expr,Nil);
  bdd_ptr goal_states = bdd_not(dd_manager, tmp);

  bdd_free(dd_manager, tmp);
  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr,"Size of invariant states: %d bdd nodes, %g states\n",
            bdd_size(dd_manager, goal_states), bdd_count_states(dd_manager, goal_states));
    fprintf(nusmv_stderr,"Size of initial states: %d bdd nodes.\n", bdd_size(dd_manager, init_bdd));
    fprintf(nusmv_stderr,"Starting to slice the space of reachable states ...\n");
  }

  plan_reach_list = cons((node_ptr)bdd_dup(f_frontier), plan_reach_list);
  reachable_states = bdd_dup(init_bdd);
  while ((f_frontier != zero) && (target_states == zero)) {
    bdd_ptr f_prev_reachable = bdd_dup(reachable_states);
    
    if (opt_verbose_level_gt(options, 0))
      fprintf(nusmv_stderr,"iteration %d: BDD size: %d bdd nodes, states = %g; frontier size = %d bdd nodes.\n",
              nstep, bdd_size(dd_manager, reachable_states),
              bdd_count_states(dd_manager, reachable_states),
              bdd_size(dd_manager, f_frontier));
    
    /* Intersect the frontier with normal assignments */
    bdd_and_accumulate(dd_manager, &f_frontier, invar_bdd);
    if (opt_verbose_level_gt(options, 0))
      fprintf(nusmv_stderr,"invariant combined: %d bdd nodes\n",
              bdd_size(dd_manager, f_frontier));
    {
      bdd_ptr not_f_prev_reachable = bdd_not(dd_manager, f_prev_reachable);
      bdd_ptr image = Img_ImageFwdMonolithicOpt(f_frontier, f_prev_reachable, not_f_prev_reachable);

      bdd_or_accumulate(dd_manager, &reachable_states, image);
      bdd_free(dd_manager, f_frontier);
      f_frontier = bdd_and(dd_manager, reachable_states, not_f_prev_reachable);
      bdd_free(dd_manager, not_f_prev_reachable);
      bdd_free(dd_manager, image);
    }
    if (opt_verbose_level_gt(options, 0))
      fprintf(nusmv_stderr, "forward step done: %d bdd nodes.\n",
              bdd_size(dd_manager, reachable_states));

    plan_reach_list = cons((node_ptr)bdd_dup(f_frontier), plan_reach_list);
    if (opt_verbose_level_gt(options, 0))
      fprintf(nusmv_stderr,"new frontier computed: %d bdd nodes.\n",
              bdd_size(dd_manager, f_frontier));
    bdd_free(dd_manager, target_states);
    target_states = bdd_and(dd_manager, goal_states, f_frontier);
    if (opt_verbose_level_gt(options, 0)) {
      if (target_states != zero) {
        fprintf(nusmv_stderr, "States not satisfying invariant reached at step %d: %d bdd nodes, %g states\n",
                nstep, bdd_size(dd_manager, target_states), bdd_count_states(dd_manager, target_states));
      }
    }
    bdd_free(dd_manager, f_prev_reachable);
    nstep++;
  }
  fprintf(nusmv_stdout, "-- invariant ");
  print_node(nusmv_stdout, inv_expr);
  if (target_states != zero) {
    fprintf(nusmv_stdout, " is false\n");
    compute_and_print_path(target_states, plan_reach_list, nstep, 0);
  }
  else { 
    fprintf(nusmv_stdout, " is true\n");
  }
  bdd_free(dd_manager, target_states);
  bdd_free(dd_manager, f_frontier);
  walk_dd(dd_manager, (void (*)())bdd_free, plan_reach_list);
  bdd_free(dd_manager, zero);
  bdd_free(dd_manager, reachable_states);
  bdd_free(dd_manager, goal_states);
  {
    node_ptr l = plan_reach_list;

    while (l) {
      node_ptr m = l;

      l = cdr(l);
      free_node(m);
    }
  }
}

/**Function********************************************************************

  Synopsis           []

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void compute_and_print_path_fb(bdd_ptr target_states,
                     bdd_ptr f_target, node_ptr f_plan_reach_list, int f_step,
                     bdd_ptr b_target, node_ptr b_plan_reach_list, int b_step)
{
  /* 
     Warning this routine is not correct. It uses the compute_and_print_path,
     but it must use an ad hoc function, since the
     compute_and_print_path generates a new trace at each call.
  */
  if (f_step != 0) compute_and_print_path(target_states, f_plan_reach_list, f_step, 0);
  if (b_step != 0) compute_and_print_path(target_states, b_plan_reach_list, b_step, 1);
}


/**Function********************************************************************

  Synopsis           [Performs on the fly verification of the
  invariant during reachability analysis.]

  Description        [During the computation of reachable states it
  checks invariants. If the invariant is not satisfied, then an
  execution trace leading to a state not satisfing the invariant is
  printed out. This function differs from check_invariant_forward
  since it performs backward and forward search.]

  SideEffects        []

  SeeAlso            [check_invariant_forward]

******************************************************************************/
void check_invariant_forward_backward(node_ptr inv_expr)
{ 
  int f_step = 0;
  int b_step = 0;
  int turn = 1;
  node_ptr f_plan_reach_list = Nil;
  node_ptr b_plan_reach_list = Nil;
  bdd_ptr zero = bdd_zero(dd_manager);
  /* computes the BDD representing the invar to check */
  bdd_ptr f_target = eval_spec(inv_expr,Nil);
  bdd_ptr b_target = bdd_dup(init_bdd);
  bdd_ptr f_cur_reachable = bdd_dup(b_target);
  bdd_ptr b_cur_reachable = bdd_dup(f_target);
  bdd_ptr f_prev_reachable = bdd_dup(f_cur_reachable);
  bdd_ptr b_prev_reachable = bdd_dup(b_cur_reachable);
  bdd_ptr f_frontier = bdd_dup(f_cur_reachable);
  bdd_ptr b_frontier = bdd_dup(b_cur_reachable);
  bdd_ptr target_states = bdd_and(dd_manager, f_target, b_target);

  f_plan_reach_list = cons((node_ptr)bdd_dup(f_frontier), f_plan_reach_list);
  b_plan_reach_list = cons((node_ptr)bdd_dup(b_frontier), b_plan_reach_list);
  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "size of target sets: forward = %d bdd nodes, backward = %d bdd nodes.\n",
	    bdd_size(dd_manager, f_target), bdd_size(dd_manager, b_target));
    fprintf(nusmv_stderr, "Starting search ...\n");
  }
  
  while ((f_frontier != zero) && (b_frontier != zero) && (target_states == zero)) {
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Iteration %d, %d forward, %d backward.\n",
              f_step + b_step, f_step, b_step);
      fprintf(nusmv_stderr, "forward frontier size = %d bdd nodes, %g states.\n",
              bdd_size(dd_manager, f_frontier),
              bdd_count_states(dd_manager, f_frontier));
      fprintf(nusmv_stderr, "backward frontier size = %d bdd nodes, %g states.\n",
              bdd_size(dd_manager, b_frontier),
              bdd_count_states(dd_manager, b_frontier));
    }
#ifndef FB_HEURISTIC
    if (FORWARD_BACKWARD_HEURISTICS == 1) { /* Step forward */
#else
    if /* (turn == 1) */
      (bdd_size(dd_manager, f_frontier) < bdd_size(dd_manager, b_frontier)) { 
      turn = 0;
#endif
      {
        bdd_ptr image = Img_ImageFwd(f_frontier);
 
        bdd_free(dd_manager, f_prev_reachable);
        f_prev_reachable = bdd_dup(f_cur_reachable);
        bdd_or_accumulate(dd_manager, &f_cur_reachable, image);
        bdd_free(dd_manager, image);
      }

      if (opt_verbose_level_gt(options, 0))
        fprintf(nusmv_stderr, "forward step done: %d bdd nodes\n",
                bdd_size(dd_manager, f_cur_reachable));
      {
        bdd_ptr not_f_prev_reachable = bdd_not(dd_manager, f_prev_reachable);

        f_frontier = bdd_and(dd_manager, f_cur_reachable, not_f_prev_reachable);
        bdd_free(dd_manager, not_f_prev_reachable);
      }

      if (opt_verbose_level_gt(options, 0))
        fprintf(nusmv_stderr,"new frontier computed: %d bdd nodes.\n",
                bdd_size(dd_manager, f_frontier));
      bdd_and_accumulate(dd_manager, &f_frontier, invar_bdd);
      if (opt_verbose_level_gt(options, 0))
        fprintf(nusmv_stderr,"frontier simplification with invariants, new size = %d bdd nodes.\n",
                bdd_size(dd_manager, f_frontier));
      bdd_free(dd_manager, b_target);
      b_target = bdd_dup(f_frontier);
      bdd_free(dd_manager, target_states);
      target_states = bdd_and(dd_manager, f_frontier, f_target);
      f_plan_reach_list = cons((node_ptr)bdd_dup(f_frontier), f_plan_reach_list);
      if (opt_verbose_level_gt(options, 0)) {
        if (target_states == zero) {
          fprintf(nusmv_stderr, "States satisfying invariant set reached at iteration %d forward, %d backward.\n", f_step, b_step);
        } else {
          fprintf(nusmv_stderr, "States not satisfying invariant set reached at step %d, size = %d bdd nodes.", f_step + b_step, bdd_size(dd_manager, target_states));
        }
      }
      f_step++;
    } else {/* Step backward */
#ifdef FB_HEURISTIC
      turn = 1;
#endif
      {
        bdd_ptr image = Img_ImageFwd(b_frontier);

        bdd_free(dd_manager, b_prev_reachable);
        b_prev_reachable = bdd_dup(b_cur_reachable);
        bdd_or_accumulate(dd_manager, &b_cur_reachable, image);
        bdd_free(dd_manager, image);
      }

      if (opt_verbose_level_gt(options, 0))
        fprintf(nusmv_stderr, "backward step done, size = %d bdd nodes.\n",
                bdd_size(dd_manager, b_cur_reachable));
      {
        bdd_ptr not_b_prev_reachable = bdd_not(dd_manager, b_prev_reachable);

        b_frontier = bdd_and(dd_manager, b_cur_reachable, not_b_prev_reachable);
        bdd_free(dd_manager, not_b_prev_reachable);
      }

      if (opt_verbose_level_gt(options, 0))
        fprintf(nusmv_stderr,"new frontier computed, size = %d bdd nodes.\n",
                bdd_size(dd_manager, b_frontier));
      bdd_and_accumulate(dd_manager, &b_frontier, invar_bdd);
      if (opt_verbose_level_gt(options, 0))
        fprintf(nusmv_stderr, "frontier simplification with invariants, new size = %d bdd nodes.\n",
                bdd_size(dd_manager, b_frontier));

      bdd_free(dd_manager, f_target);
      f_target = bdd_dup(b_frontier);
      bdd_free(dd_manager, target_states);
      target_states = bdd_and(dd_manager, b_frontier, b_target);
      b_plan_reach_list = cons((node_ptr)bdd_dup(b_frontier), b_plan_reach_list);
      if (opt_verbose_level_gt(options, 0)) {
        if (target_states == zero) {
          fprintf(nusmv_stderr, "States satisfying invariant set reached at iteration %d forward, %d backward\n", f_step, b_step);
        } else {
          fprintf(nusmv_stderr, "States not satisfying invariant set reached at step %d, size = %d bdd nodes.\n", f_step + b_step, bdd_size(dd_manager, target_states));
        }
      }
      b_step++;
    }
  }
  if ((f_frontier == 0) || (b_frontier == zero)) {
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "\n****************************************\n");
      fprintf(nusmv_stderr, "The invariant is satisfied.\n");
      fprintf(nusmv_stderr, "****************************************\n");
    }
  } else {
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "\n****************************************\n");
      fprintf(nusmv_stderr, "A counterexample is:\n");
      fprintf(nusmv_stderr, "****************************************\n");
    }
    compute_and_print_path_fb(target_states, f_target, f_plan_reach_list, f_step, b_target, b_plan_reach_list, b_step);
  }
  walk_dd(dd_manager, (void (*)())bdd_free, f_plan_reach_list);
  walk_dd(dd_manager, (void (*)())bdd_free, b_plan_reach_list);
  
  bdd_free(dd_manager, f_target);
  bdd_free(dd_manager, b_target);
  bdd_free(dd_manager, f_cur_reachable);
  bdd_free(dd_manager, b_cur_reachable);
  bdd_free(dd_manager, f_prev_reachable);
  bdd_free(dd_manager, b_prev_reachable);
  bdd_free(dd_manager, f_frontier);
  bdd_free(dd_manager, b_frontier);
  bdd_free(dd_manager, target_states);  
  bdd_free(dd_manager, zero);
  {
    node_ptr l = f_plan_reach_list;

    while (l) {
      node_ptr m = l;

      l = cdr(l);
      free_node(m);
    }
  }
  {
    node_ptr l = b_plan_reach_list;

    while (l) {
      node_ptr m = l;

      l = cdr(l);
      free_node(m);
    }
  }
}

/**Function********************************************************************

  Synopsis           [Performs on the fly verification of the
  invariant during reachability analysis.]

  Description        [During the computation of reachable states it
  checks invariants. If the invariant is not satisfied, then an
  execution trace leading to a state not satisfing the invariant is
  printed out. This function differs from check_invariant_forward_fb
  since it uses an optimized algorithm for image computation based on
  don't care sets. It works only with the monolithic transition relation.]

  SideEffects        []

  SeeAlso            [check_invariant_forward]

******************************************************************************/
void check_invariant_forward_backward_opt(node_ptr inv_expr)
{ 
  int f_step = 0;
  int b_step = 0;
  int turn = 1;
  node_ptr f_plan_reach_list = Nil;
  node_ptr b_plan_reach_list = Nil;
  bdd_ptr zero = bdd_zero(dd_manager);
  /* computes the BDD representing the invar to check */
  bdd_ptr f_target = eval_spec(inv_expr,Nil);
  bdd_ptr b_target = bdd_dup(init_bdd);
  bdd_ptr f_cur_reachable = bdd_dup(b_target);
  bdd_ptr b_cur_reachable = bdd_dup(f_target);
  bdd_ptr f_prev_reachable = bdd_dup(f_cur_reachable);
  bdd_ptr b_prev_reachable = bdd_dup(b_cur_reachable);
  bdd_ptr f_frontier = bdd_dup(f_cur_reachable);
  bdd_ptr b_frontier = bdd_dup(b_cur_reachable);
  bdd_ptr target_states = bdd_and(dd_manager, f_target, b_target);

  f_plan_reach_list = cons((node_ptr)bdd_dup(f_frontier), f_plan_reach_list);
  b_plan_reach_list = cons((node_ptr)bdd_dup(b_frontier), b_plan_reach_list);
  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "Size of target sets: forward = %d bdd nodes, backward = %d bdd nodes.\n",
	    bdd_size(dd_manager, f_target), bdd_size(dd_manager, b_target));
    fprintf(nusmv_stderr, "Starting search ...\n");
  }
  
  while ((f_frontier != zero) && (b_frontier != zero) && (target_states == zero)) {
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Iteration %d, %d forward, %d backward.\n",
              f_step + b_step, f_step, b_step);
      fprintf(nusmv_stderr, "forward frontier size = %d bdd nodes, %g states.\n",
              bdd_size(dd_manager, f_frontier),
              bdd_count_states(dd_manager, f_frontier));
      fprintf(nusmv_stderr, "backward frontier size = %d bdd nodes, %g states.\n",
              bdd_size(dd_manager, b_frontier),
              bdd_count_states(dd_manager, b_frontier));
    }
#ifndef FB_HEURISTIC
    if (FORWARD_BACKWARD_HEURISTICS == 1) { /* Step forward */
#else
    if /* (turn == 1) { */
      (bdd_size(dd_manager, f_frontier) < bdd_size(dd_manager, b_frontier)) { 
      turn = 0;
#endif
      {
        bdd_ptr not_f_prev_reachable, image;
 
        bdd_free(dd_manager, f_prev_reachable);
        f_prev_reachable = bdd_dup(f_cur_reachable);
        not_f_prev_reachable = bdd_not(dd_manager, f_prev_reachable);
        image = Img_ImageFwdMonolithicOpt(f_frontier, f_prev_reachable, not_f_prev_reachable); 
        bdd_or_accumulate(dd_manager, &f_cur_reachable, image);
        bdd_free(dd_manager, f_frontier);
        f_frontier = bdd_and(dd_manager, f_cur_reachable, not_f_prev_reachable);
        bdd_free(dd_manager, not_f_prev_reachable);
        bdd_free(dd_manager, image);
      }

      if (opt_verbose_level_gt(options, 0)) fprintf(nusmv_stderr, "forward step done, size = %d bdd nodes.\n",
                           bdd_size(dd_manager, f_cur_reachable));
      if (opt_verbose_level_gt(options, 0)) fprintf(nusmv_stderr,"new frontier computed, size = %d bdd nodes.\n",
                           bdd_size(dd_manager, f_frontier));
      bdd_and_accumulate(dd_manager, &f_frontier, invar_bdd);
      if (opt_verbose_level_gt(options, 0)) fprintf(nusmv_stderr,"frontier simplification with invariants, new size = %d bdd nodes.\n",
                           bdd_size(dd_manager, f_frontier));
      bdd_free(dd_manager, b_target);
      b_target = bdd_dup(f_frontier);
      bdd_free(dd_manager, target_states);
      target_states = bdd_and(dd_manager, f_frontier, f_target);
      f_plan_reach_list = cons((node_ptr)bdd_dup(f_frontier), f_plan_reach_list);
      if (opt_verbose_level_gt(options, 0)) {
        if (target_states == zero) {
          fprintf(nusmv_stderr, "Goal not reached at iteration %d forward, %d backward.\n", f_step, b_step);
        } else {
          fprintf(nusmv_stderr, "Goal set reached at step %d, size = %d bdd nodes.", f_step + b_step,
                  bdd_size(dd_manager, target_states));
        }
      }
      f_step++;
    } else {/* Step backward */
#ifdef FB_HEURISTIC
      turn = 1;
#endif
      {
        bdd_ptr not_b_prev_reachable, image;

        bdd_free(dd_manager, b_prev_reachable);
        b_prev_reachable = bdd_dup(b_cur_reachable);
        not_b_prev_reachable = bdd_not(dd_manager, b_prev_reachable);
        image = Img_ImageFwdMonolithicOpt(b_frontier, b_prev_reachable, not_b_prev_reachable);
        bdd_or_accumulate(dd_manager, &b_cur_reachable, image);
        bdd_free(dd_manager, b_frontier);
        b_frontier = bdd_and(dd_manager, b_cur_reachable, not_b_prev_reachable);
        bdd_free(dd_manager, not_b_prev_reachable);
        bdd_free(dd_manager, image);
      }

      if (opt_verbose_level_gt(options, 0)) {
        fprintf(nusmv_stderr, "backward step done, size = %d bdd nodes.\n",
                bdd_size(dd_manager, b_cur_reachable));
        fprintf(nusmv_stderr,"new frontier computed, size = %d bdd nodes. \n",
                bdd_size(dd_manager, b_frontier));
      }
      bdd_and_accumulate(dd_manager, &b_frontier, invar_bdd);
      if (opt_verbose_level_gt(options, 0))
        fprintf(nusmv_stderr,"frontier simplification with invariants, new size = %d bdd nodes.\n",
                bdd_size(dd_manager, b_frontier));

      bdd_free(dd_manager, f_target);
      f_target = bdd_dup(b_frontier);
      bdd_free(dd_manager, target_states);
      target_states = bdd_and(dd_manager, b_frontier, b_target);
      b_plan_reach_list = cons((node_ptr)bdd_dup(b_frontier), b_plan_reach_list);
      if (opt_verbose_level_gt(options, 0)) {
        if (target_states == zero) {
          fprintf(nusmv_stderr, "Goal not reached at iteration %d forward, %d backward\n", f_step, b_step);
        } else {
          fprintf(nusmv_stderr, "Goal set reached at step %d, size = %d bdd nodes.\n", f_step + b_step,
                  bdd_size(dd_manager, target_states));
        }
      }
      b_step++;
    }
  }
  if ((f_frontier == 0) || (b_frontier == zero)) {
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "\n****************************************\n");
      fprintf(nusmv_stderr, "Fix point found, goal is unreachable\n");
      fprintf(nusmv_stderr, "****************************************\n");
    }
  } else {
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "\n****************************************\n");
      fprintf(nusmv_stderr, "The plan is:\n");
      fprintf(nusmv_stderr, "****************************************\n");
    }
    compute_and_print_path_fb(target_states, f_target, f_plan_reach_list, f_step, b_target, b_plan_reach_list, b_step);
  }
  walk_dd(dd_manager, (void (*)())bdd_free, f_plan_reach_list);
  walk_dd(dd_manager, (void (*)())bdd_free, b_plan_reach_list);
  
  bdd_free(dd_manager, f_target);
  bdd_free(dd_manager, b_target);
  bdd_free(dd_manager, f_cur_reachable);
  bdd_free(dd_manager, b_cur_reachable);
  bdd_free(dd_manager, f_prev_reachable);
  bdd_free(dd_manager, b_prev_reachable);
  bdd_free(dd_manager, f_frontier);
  bdd_free(dd_manager, b_frontier);
  bdd_free(dd_manager, target_states);  
  bdd_free(dd_manager, zero);
  {
    node_ptr l = f_plan_reach_list;

    while (l) {
      node_ptr m = l;

      l = cdr(l);
      free_node(m);
    }
  }
  {
    node_ptr l = b_plan_reach_list;

    while (l) {
      node_ptr m = l;

      l = cdr(l);
      free_node(m);
    }
  }
}


/**Function********************************************************************

  Synopsis           [Print an invariant specification]

  Description        [Print an invariant specification]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void print_invar(FILE *file, node_ptr n)
{
  node_ptr context = Nil;
  
  if (node_get_type(n) == CONTEXT) {
    context = car(n);
    n = cdr(n);
  }
  indent_node(file, "invariant ", n, " ");
  if (context) {
    fprintf(file, "(in module ");
    print_node(file, context);
    fprintf(file, ") ");
  }
}
