/**CFile***********************************************************************

  FileName    [simulate.c]

  PackageName [simulate]

  Synopsis    [Simulator routines]

  Description [This file contains functions for image computation, for state
	picking in a set (according to three different policies), for states display
	and constraints insertion.]

  Author      [Andrea Morichetti]

  Copyright   [Copyright (c) 1998 by ITC-IRST and Carnegie Mellon University.
  All Rights Reserved. This software is for educational purposes only.
  Permission is given to academic institutions to use, copy, and modify
  this software and its documentation provided that this introductory
  message is not removed, that this software and its documentation is
  used for the institutions' internal research and educational purposes,
  and that no monies are exchanged. No guarantee is expressed or implied
  by the distribution of this code.
  Send bug-reports and/or questions to: nusmv@irst.itc.it.]

******************************************************************************/

#include "simulateInt.h"

static char rcsid[] UTIL_UNUSED = "$Id: $";

/**Variable********************************************************************

  Synopsis    [Stack and context for non-local goto inside simulator]

  Description [Stack and context for non-local goto inside simulator]

******************************************************************************/
#if SOLARIS
static sigjmp_buf simulate_jmp_buff;
#else
static jmp_buf simulate_jmp_buff;
#endif

static  void (*saved_simulate_sigterm)(int);
/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static bdd_ptr pick_state_rnd_or_det ARGS((bdd_ptr, Simulation_Mode));
static void show_next_states ARGS((bdd_ptr *, int, int));
static void simulate_sigterm ARGS((int sig));
			  
/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/
/**Function********************************************************************

  Synopsis           [Signal handler]

  Description        [SIGINT signal handler inside the simulator.]

  SideEffects        []

******************************************************************************/
static void simulate_sigterm(int sig) {
  signal(SIGINT, saved_simulate_sigterm);
  LONGJMP(simulate_jmp_buff, 1);
}

/**Function********************************************************************

  Synopsis           [Extracts a minterm from a given BDD]

  Description        [Picks a state from a set according to the Simulation_Mode
  (only Random or Deterministic: Interactive mode is not supported) entered: the
  first one sets to true those boolean variables which are indifferent for the
  actual BDD; the second one chooses those boolean values non deterministically.
  This is necessary in order to have unambiguous assignments for BDDs boolean
  variables.]

  SideEffects        []

  SeeAlso            [bdd_pick_one_state_rand bdd_pick_one_state]

******************************************************************************/
static bdd_ptr pick_state_rnd_or_det(bdd_ptr s, Simulation_Mode mode)
{
  if (mode != Random && mode != Deterministic) {
    fprintf(nusmv_stderr, "Warning: (pick_state_rnd_or_det) \"Deterministic\" mode will be used.");
    mode = Deterministic;
  }
  return(mode == Random ? bdd_pick_one_state_rand(dd_manager,s) :
	 bdd_pick_one_state(dd_manager,s));
}

/**Function********************************************************************

  Synopsis           [Stores a trace in the hash table and prints it.]

  Description        [Stores a trace in the hash table and prints it if the
  variable printyesno is true (this is set by the user via the command simulate
  options -p or -v).]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void store_and_print_trace(node_ptr p, boolean printyesno, int verbose)
{
  node_ptr state_no = Nil;
  node_ptr trace_no = Nil;
  node_ptr current_state_label = current_state_label_get();
  boolean lengthen = false;

  if (opt_verbose_level_gt(options, 0) && printyesno == true) {
    fprintf(nusmv_stdout,
      "#####################################################\n");
    fprintf(nusmv_stdout,
      "######         Print Of Current Trace         #######\n");
    fprintf(nusmv_stdout,
      "#####################################################\n");
  }
  clear_print_hash();
  /* checks if current_state is the last one in an existent trace: if the
    case, lengthens the old trace with the current one and ignore the first
    state of the list received (because it already exists in the old trace) */
  if (lookup_traces_hash(find_node(DOT, car(current_state_label),
          node_plus1(cdr(current_state_label)))) == Nil) {
    state_no = node_plus1(cdr(current_state_label));
    trace_no = car(current_state_label);
    p = cdr(p);
    lengthen = true;
  }
  else {
    state_no = find_node(NUMBER, (node_ptr)1, Nil);
    trace_no = find_node(NUMBER, (node_ptr)get_trace_number(), Nil);
  }
  if (opt_gui_behavior(options) == false) set_indent_size(2);
  while (p){
    node_ptr label = find_node(DOT, trace_no, state_no);

    bdd_ref((bdd_ptr)car(p));
    insert_traces_hash(label, car(p));
    if (printyesno == true) {
      fprintf(nusmv_stdout, "-> State ");
      print_node(nusmv_stdout, label);
      fprintf(nusmv_stdout, " <-\n");
      print_state((bdd_ptr)car(p), verbose);
    }
    state_no = node_plus1(state_no);
    p = cdr(p);
  }
  if (lengthen == false) inc_trace_number();
  if (opt_gui_behavior(options) == false) reset_indent_size();
}

/**Function********************************************************************

  Synopsis           [Multiple step simulation]

  Description        [Multiple step simulation: loops n times over the choice of
  a state according to the picking policy given at command line. It returns a
  list of at least n+1 referenced states (the first one is always the "current
  state" from which any simulation must start). The obtained list can contain
  a minor number of states if there are no future states at some point.]

  SideEffects        []

  SeeAlso            [Simulate_ChooseOneState]

******************************************************************************/
node_ptr Simulate_MultipleSteps(bdd_ptr constraint, Simulation_Mode mode,
        int n, int display_all)
{
  bdd_ptr current_state = current_state_bdd_get();
  node_ptr result = Nil;
  int i = 1;

  /* we append the current state in first position of the trace */
  result = cons((node_ptr)current_state, result);
  while(i <= n) {
    bdd_ptr next_constr_set = (bdd_ptr)NULL;

    if (opt_verbose_level_gt(options, 0)) {
      switch (mode) {
      case Interactive:
        fprintf(nusmv_stderr,
                "********  Interactive mode Simulation step %d  *********\n",i);
        break;
      case Random:
	fprintf(nusmv_stderr,
		"**********  Random mode Simulation step %d  **********\n",i);
        break;
      case Deterministic:
	fprintf(nusmv_stderr,
		"*******  Deterministic mode Simulation step %d  *******\n",i);
        break;
      }
    }
    next_constr_set = Img_ImageFwd(current_state);
    bdd_and_accumulate(dd_manager, &next_constr_set, constraint);
    bdd_and_accumulate(dd_manager, &next_constr_set, invar_bdd);

    if (bdd_is_zero(dd_manager, next_constr_set)) {
      fprintf(nusmv_stderr, "No future state exists");
      fprintf(nusmv_stderr, (llenght(result) == 1 ? ": trace not built.\n" : "."));
      fprintf(nusmv_stderr, " Simulation stopped at step %d.\n", i);
      bdd_free(dd_manager, next_constr_set);
      result = reverse(result);
      /* We don't free the current_state variable because the
	 list "result" must contain referenced states */
      return(result);
    }

    current_state = Simulate_ChooseOneState(next_constr_set, mode, display_all);

    if (current_state == (bdd_ptr)NULL || bdd_is_zero(dd_manager, current_state)) {
      fprintf(nusmv_stderr,
        "\nCan't find a future state. Simulation stopped at step %d.\n", i);
      if (current_state != (bdd_ptr)NULL) bdd_free(dd_manager, current_state);
      bdd_free(dd_manager, next_constr_set);
      result = reverse(result);
      return(result);
    }
    result = cons((node_ptr)current_state, result);
    /* we don't free the current_state variable because the final list "result"
       must contain referenced states: they will be freed after their insertion
       in the traces_hash table */
    i++;
    bdd_free(dd_manager, next_constr_set);
  }
  result = reverse(result);
  return(result);
}

/**Function********************************************************************

  Synopsis           [Chooses one state among future states]

  Description        [Chooses a state among future states depending on the
  given simulation policy (random, deterministic or interactive). In case of
	interactive simulation, the system stops and allows the user to pick
	a state	from a list of possible items. If the number of future states
	is too high, the system requires some further constraints to limit that
	number and will asks for them until the number of states is lower than
	an internal threshold. Entered expressions are accumulated in one big
	constraint used only in the actual step of the simulation. It will be
	discarded after a state will be chosen.]

  SideEffects        [A referenced state (BDD) is returned. NULL if failed.]

  SeeAlso            [Simulate_MultipleStep]

******************************************************************************/
bdd_ptr Simulate_ChooseOneState(bdd_ptr next_set, Simulation_Mode mode,
        int display_all)
{
  bdd_ptr result = (bdd_ptr)NULL;

  saved_simulate_sigterm = signal(SIGINT, simulate_sigterm);
  if (SETJMP(simulate_jmp_buff, 1) == 0) {
    switch(mode) {
    case Interactive: {
      boolean further = false;
      int i = 0;
      int choice = -1;
      int array_of_states_size =
	bdd_count_minterm(dd_manager, next_set, state_vars_dim);
      bdd_ptr *array_of_states     = (bdd_ptr *)NULL;
      bdd_ptr  further_constraints = bdd_one(dd_manager);
      node_ptr parsed_command      = Nil;
      bdd_ptr tmp_set = (bdd_ptr)NULL;
      
      /*
	There are 4 condition to be verified in order to accept new further
	constraints:
	
	1) entered expression must be a non-zero set;
	2) entered expression must be consistent with the accumulated constraints
	(i.e. the product (further /\ accumulated) must be non-zero;
	3) if (further /\ accumulated) is non-zero, it also must be non-zero the product
	(further /\ accumulated) /\ next_set of states
	4) cardinality of the set obtained from the last product must be < shown_states
	*/
      
      while (array_of_states_size > opt_shown_states_level(options)) {
	boolean done = false;
	
	further = true;
	fprintf(nusmv_stdout,
		"Too many (%d) future states to visualize. Please specify\n",
		array_of_states_size);
	
	while(done != true) {
	  char *strConstr[2];
	  char str[CONSTR_LENGTH];
	  
	  for (i=0; i<CONSTR_LENGTH; i++) str[i] = '\0';
	  fprintf(nusmv_stdout, "further constraints: ");
	  if (opt_gui_behavior(options) == true) fprintf(nusmv_stdout, "\n");
	  
	  /* Continue if receive a "good" string */
	  if ((fgets(str, CONSTR_LENGTH, nusmv_stdin) != (char)NULL) ||
	      (str[0] != '\n')) {
	    strConstr[0] = NIL(char);
	    strConstr[1] = util_strsav(str);
	    
	    CATCH {        
	      if (Parser_ReadFromString(2,strConstr,"CONSTRAINT ",";\n",
					&parsed_command)==0) {
		if ((parsed_command != Nil) &&
		    (node_get_type(parsed_command) == CONSTRAINT)) {
		  node_ptr constraints    = car(parsed_command);
		  add_ptr add_constraints = eval(constraints, Nil);
		  
		  bdd_ptr read_constr = add_to_bdd(dd_manager, add_constraints);
		  add_free(dd_manager, add_constraints);
		  if (bdd_is_zero(dd_manager, read_constr)) {
		    fprintf(nusmv_stderr,
			    "Entered constraints are equivalent to the empty set. Try again.\n");
		    bdd_free(dd_manager, read_constr);
		  }
		  else {/* entered expression is non-zero */
		    bdd_ptr tmp_constr = bdd_and(dd_manager, further_constraints, read_constr);
		    
		    bdd_free(dd_manager, read_constr);
		    if (bdd_is_zero(dd_manager, tmp_constr)) {
		      fprintf(nusmv_stderr,
			      "Entered expression is incompatible with old constraints. Try again.\n");
		      bdd_free(dd_manager, tmp_constr);
		    }
		    else {/* total constraints are non-zero */
		      tmp_set = bdd_and(dd_manager, tmp_constr, next_set);
		      
		      if (bdd_is_zero(dd_manager, tmp_set)) {
			fprintf(nusmv_stderr,
				"Set of future states is EMPTY: constraints too strong ? Try again.\n");
			bdd_free(dd_manager, tmp_constr);
			bdd_free(dd_manager, tmp_set);
		      }
		      else {/* next restricted set of states is non-zero */
			bdd_and_accumulate(dd_manager, &further_constraints, tmp_constr);
			
			bdd_free(dd_manager, tmp_constr);
			done = true;
			array_of_states_size =
			  bdd_count_minterm(dd_manager, tmp_set, state_vars_dim);
		      }
		    }
		  }
		}/* end if(parsed_command) */
		else
		  fprintf(nusmv_stderr,
			  "Parsing error: constraints must be \"simple_expressions\".\n");
	      } /* end if(Parser_ReadFromString) */
	      else
		fprintf(nusmv_stderr,
			"Parsing error: constraints must be \"simple_expressions\".\n");
	    } /* end CATCH */
	  } /* end if (fgets) */
	} /* end while (done != true)*/
      } /* end while (array_of_states_size) */
      bdd_free(dd_manager, further_constraints);
      array_of_states = ALLOC(bdd_ptr, array_of_states_size);
      assert(array_of_states == (bdd_ptr *)NULL);
      
      { /* chooses the set to pick from and fills the array */
	int res = bdd_pick_all_terms(dd_manager, ((tmp_set != (bdd_ptr)NULL) ?
						  tmp_set : next_set), state_vars, state_vars_dim,
				     array_of_states, array_of_states_size);
	
	if (tmp_set != (bdd_ptr)NULL) bdd_free(dd_manager, tmp_set);
	if (res == 1) {
	  FREE(array_of_states);
	  return((bdd_ptr)NULL);
	}
      }
      show_next_states(array_of_states, array_of_states_size, display_all);
      
      /* Input of the choice */
      if (array_of_states_size > 1) {
	char line[CHOICE_LENGTH];
	
	for (i=0; i<CHOICE_LENGTH; i++) line[i] = '\0';
	fprintf(nusmv_stdout,
		"\nChoose a state from the above (0-%d): ",array_of_states_size-1);
	if (opt_gui_behavior(options) == true) fprintf(nusmv_stdout, "\n");
	
	while((fgets(line, CONSTR_LENGTH, nusmv_stdin) != (char)NULL) ||
	      (line[0] != '\n')) {
	  if((sscanf(line, "%d", &choice) != 1) || (choice < 0) ||
	     (choice>array_of_states_size-1)) {
	    fprintf(nusmv_stdout,"Choose a state from the above (0-%d): ",
		    array_of_states_size-1);
	    if (opt_gui_behavior(options) == true) fprintf(nusmv_stdout, "\n");
	    continue;
	  }
	  else break;
	}
      }
      else {
	fprintf(nusmv_stdout,
		"\nThere's only one future state. Press Return to Proceed.");
	if (opt_gui_behavior(options) == true) fprintf(nusmv_stdout, "\n");
	while ((choice = getc(nusmv_stdin)) != '\n') {;}
	if (opt_gui_behavior(options) == true) fprintf(nusmv_stdout, "\n");
	choice = 0;
      }
      fprintf(nusmv_stdout, "\nChoosen state is: %d\n", choice);
      print_state(array_of_states[choice], display_all);
      fprintf(nusmv_stdout, "\n");
      if (opt_gui_behavior(options) == true) fprintf(nusmv_stdout, "\n");
      result = bdd_dup(array_of_states[choice]);
      /* 
         The array returned by bdd_pick_all_terms has ref greater than 0.
         We have to dereference them
      */
      for (i=0; i < array_of_states_size; i++)
	bdd_free(dd_manager,array_of_states[i]);
      FREE(array_of_states);
      break;
    }
    default:
      result = pick_state_rnd_or_det(next_set, mode);
      break;
    } 
  } else {
    (void) fprintf(nusmv_stderr, "\n");
    /* signal(SIGINT, restore_signal); */
  }
  return(result);
}

/**Function********************************************************************

  Synopsis           [Prints an array of states on nusmv_stdout.]

  Description        [Prints an array of states on nusmv_stdout. If
  <tt>changes_only</tt> is 1, then only state variables which assume a
  different value from the previous printed one will be printed out.]

  SideEffects        []

******************************************************************************/
void show_next_states(bdd_ptr *array, int size, int changes_only)
{
  int j = 0;

  fprintf(nusmv_stdout,
    "***************  AVAILABLE FUTURE STATES  *************\n");
  clear_print_hash();
  for(j=0; j < size; j++){
    if (opt_gui_behavior(options) == false) fprintf(nusmv_stdout, "\n");
    fprintf(nusmv_stdout, "%d) -------------------------\n", j);
    set_indent_size(2);
    print_state(array[j], changes_only);
    reset_indent_size();
  }
}

