/**CFile***********************************************************************

  FileName    [smMisc.c]

  PackageName [sm]

  Synopsis    [This file contain the main routine for the batch use of NuSMV.]

  Description [This file contain the main routine for the batch use of
  NuSMV. The batch main executes the various model checking steps in a
  predefined order. After the processing of the input file than it
  return to the calling shell.]

  SeeAlso     []

  Author      [Marco Roveri]

  Copyright   [ Copyright (c) 1998 by ITC-IRST and Carnegie Mellon University.
  All Rights Reserved. This software is for educational purposes only.
  Permission is given to academic institutions to use, copy, and modify
  this software and its documentation provided that this introductory
  message is not removed, that this software and its documentation is
  used for the institutions' internal research and educational purposes,
  and that no monies are exchanged. No guarantee is expressed or implied
  by the distribution of this code.
  Send bug-reports and/or questions to: nusmv@irst.itc.it. ]

******************************************************************************/

#include "smInt.h"

static char rcsid[] UTIL_UNUSED = "$Id: smMisc.c,v$";

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [The batch main.]

  Description [The batch main. It first read the input file, than
  flatten the hierarchy. Aftre this preliminar phase it creates the
  boolean variables necessary for the encoding and then start
  compiling the read model into BDD. Now computes the reachable states
  depending if the flag has been set. before starting verifying if the
  properties specified in the model hold or not it computes the
  fairness constraints. You can also activate the reordering and
  also choose to avoid the verification of the properties.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
void smBatchMain()
{
  node_ptr init_expr         = Nil;
  node_ptr invar_expr        = Nil;
  node_ptr trans_expr        = Nil;
  node_ptr assign_expr       = Nil;
  node_ptr procs_expr        = Nil;
  node_ptr fair_expr         = Nil;
  node_ptr spec_expr         = Nil;
  node_ptr ltlspec_expr      = Nil;
  node_ptr invar_spec_expr   = Nil;

  /* Necessary to have standard behavior in the batch mode */
  cancel_util_setjmp();

  /* Parse the input file */
  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "Parsing file %s ..... ", get_input_file(options));
    fflush(nusmv_stderr);
  }

  Parser_OpenInput(get_input_file(options));

  if (yyparse()) nusmv_exit(1);

  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "done.\n");
    fflush(nusmv_stderr);
  }
  
  Parser_CloseInput();
  parse_tree = reverse(parse_tree);

  /* Initializes the module hash */
  init_module_hash();
  /* check_all_implements(parse_tree); */

  /*
    Processing of the parse tree and constructions of all the
    expressions for the state machine(s). Here the expansions are
    performed so that modules and processes are created. The expansion
    of modules is such that the formal parameters (if any) are
    replaced by the actual ones and the machine is replicated.
  */
  Compile_FlattenHierarchy(sym_intern("main"), Nil, &trans_expr, &init_expr, &invar_expr,
                  &spec_expr, &ltlspec_expr, &invar_spec_expr, &fair_expr, &assign_expr,
                  &procs_expr, Nil);

  all_symbols = reverse(all_symbols);
  state_variables   = reverse(state_variables);
  input_variables = reverse(input_variables);
  /* We want input variables at the top of the ordering as default */
  all_variables = append_ns(input_variables, state_variables);
  fair_expr     = reverse(fair_expr);
  spec_expr     = reverse(spec_expr);
  ltlspec_expr  = reverse(ltlspec_expr);
  invar_spec_expr    = reverse(invar_spec_expr);

  /* Initializes the build_vars routines */
  Compile_EncodeVarsInit();
  /*
    Builds the ADD containing the list of state variables
    (i.e. vars).
    Side effect on symbol_hash.
  */
  create_process_symbolic_variables(procs_expr);
  Compile_BuildVarsBdd();

  /* Builds the ADD list representing real_state_variables */
  build_real_state_variables();

  /*
    Builds the ADD containing the list of state variables.
    Side effect on:
    - process_selector_add
    - symbol_hash (adds all <process>.running ADD leaves)
  */
  build_proc_selector(procs_expr);

  /* Builds the ADD representing frame-axioms */
  Compile_CompileFrame();

  /*
    Checking for the correctness of the state machines which have
    been declared. It should not mention DD manipulation.
  */
  Compile_CheckProgram(procs_expr, spec_expr, ltlspec_expr, invar_spec_expr, fair_expr);

  /*
    Builds the ADD representing the Initial States. This is obtained
    by conjoining the init in different modules and processes.
  */
  Compile_CompileInit(init_expr, procs_expr);

  /* Builds the ADD representing the Transition Relation */
  {
    add_ptr one = add_one(dd_manager);

    Compile_CompileModel(trans_expr, invar_expr, procs_expr, one);
    add_free(dd_manager, one);
  }

  /* Builds the OBDD representing Reachable States Set */
  if (opt_forward_search(options)) {
    if (opt_verbose_level_gt(options, 0))
      fprintf(nusmv_stderr,"Computing reachable states ...\n");
    if (opt_verbose_level_gt(options, 0))
      fprintf(nusmv_stderr,"...done\n");
    compute_reachable_states();
  }

  /* Builds the OBDD representing the Fairness constraints */
  if (opt_verbose_level_gt(options, 0))
    fprintf(nusmv_stderr,"evaluating fairness constraints...\n");
  compute_fairness_constraints(fair_expr);
  if (opt_verbose_level_gt(options, 0))
    fprintf(nusmv_stderr,"...done\n");

  if (opt_check_trans(options)) {
    check_transition_relation();
  }
  
  /* Evaluates the Specifications */
  if (!opt_ignore_spec(options)) {
    node_ptr l = spec_expr;

    while (l) { /* Loop over specifications */
      node_ptr spec = car(l);

      l = cdr(l);
      if (opt_verbose_level_gt(options, 0)) {
        fprintf(nusmv_stderr, "evaluating ");
        print_spec(nusmv_stderr, spec);
        fprintf(nusmv_stderr, "\n");
      }
      if(!(opt_ag_only(options) && opt_forward_search(options))) {
        node_ptr actual_spec = cdr(spec);

        /* decide if it is a SPEC or a COMPUTE */
        switch (node_get_type(actual_spec)) {
        case MINU:
        case MAXU: check_compute(spec); break;
        default:   check_spec(spec);    break;
        };
      }
      else
        check_AG_only(spec, Nil);
    }
  }

  /* Evaluates the LTL specifications */
  if (!opt_ignore_ltlspec(options)) {
    node_ptr l = ltlspec_expr;

    while(l) { /* Loop over specifications */
      node_ptr spec = car(l);

      l = cdr(l);
      if (opt_verbose_level_gt(options, 0)) {
        fprintf(nusmv_stderr, "evaluating ");
        print_ltlspec(nusmv_stderr, spec);
        fprintf(nusmv_stderr, "\n");
      }
      check_ltlspec(spec);
    }
  }
  /* Evaluates CHECKINVARIANTS */
  if (!opt_ignore_invar(options)) {
    node_ptr l = invar_spec_expr;

    while ( l ) {
      node_ptr invar = car(l);

      l = cdr(l);
      if (opt_verbose_level_gt(options, 0)) {
        fprintf(nusmv_stderr, "evaluating ");
        print_invar(nusmv_stderr, invar);
        fprintf(nusmv_stderr, "\n");
      }
      check_invar(invar);
    }
  }
  /* Reporting of statistical information. */
  print_usage();

  /* Computing and Reporting of the Effect of Reordering */
  if (opt_reorder(options)) {
    fprintf(nusmv_stdout, "\n========= starting reordering ============\n");
    dd_reorder(dd_manager, get_reorder_method(options), DEFAULT_MINSIZE);
    Compile_WriteOrder(get_output_order_file(options), 0);
    fprintf(nusmv_stdout, "\n========= after reordering ============\n");
    print_usage();
  }
  /* Reporting of Reachable States */
  print_reachable_states();
}

/**Function********************************************************************

  Synopsis           [Prints usage statistic.]

  Description        [Prints on <code>nusmv_stdout</code> usage
  statistics, i.e. the amount of memory used, the amount of time
  spent, the number of BDD nodes allocated and the size of the model
  in BDD.]

  SideEffects        []

  SeeAlso            [print_model_statistic]
******************************************************************************/
void print_usage(void)
{
  fprintf(nusmv_stdout, "######################################################################\n");
  util_print_cpu_stats(nusmv_stdout);
  fprintf(nusmv_stdout, "######################################################################\n");
  fprintf(nusmv_stdout, "BDD statistics\n");
  fprintf(nusmv_stdout, "--------------------\n");
  fprintf(nusmv_stdout, "BDD nodes allocated: %d\n", get_dd_nodes_allocated(dd_manager));
  print_model_statistic();
}

